import json
from pyhectiqlab.events_manager import EventsManager

class Config():
    """
    A pyhectiqlab.Config object is an abstract object designed to store 
    your configurations with a simple API.
    
    """
    def __init__(self, **kwargs):
        self._state = {}
        self._lock = False

        for key in kwargs:
            if key=="dict":
                raise KeyError("Key dict is a protected attribute.")
            if key=="keys":
                raise KeyError("Key keys is a protected attribute.")
            setattr(self, key, kwargs[key])

    def save(self, path):
        json.dump(self.dict, open(path, "w"))
    
    @classmethod
    def load(cls, path):
        return cls.from_dict(json.load(open(path, "r")))
    
    @staticmethod
    def download(run: str, project: str):
        """
        Download configuration from a run.

        run: Slug path to the run.
        project: Name of the project.
        """
        events_manager = EventsManager()
        result = events_manager.add_event("fetch_run_configs", (run, project), async_method=False)
        return Config._convert_json_res(result)

    def lock(self):
        """Freeze the state of the config.
        """
        self._lock = True

    def unlock(self):
        """Unfreeze the state of the config.
        """
        self._lock = False
    
    @staticmethod
    def from_dict(data):
        """Convert a dict to a Config object.
        `data` can be nested (dict within dict), which will generate sub configs.
        If `data` is not a dict, then the method returns `data`.

        Example:
        -------------
        d = {"a": {"b": 4}}
        config = Config.from_dict(d)
        assert config.a.b == 4
        """
        if isinstance(data, dict)==False:
            return data
        config = Config()
        for key in data:
            if isinstance(data[key], dict):
                config[key] = Config.from_dict(data[key])
            else:
                config[key] = data[key]
        return config

    @staticmethod
    def _convert_json_res(res: dict):
        """Convert the download config, meta into a Config.
        """
        meta = {}
        config = Config()
        
        if res is None:
            return Config(), {}
        
        for key, item in res.get("__meta__", {}).items():
            meta[key] = item

        config = Config.from_dict(res.get("__config__", {}))
        return config, meta

    @property
    def dict(self):
        """A property that converts a config to a dict. Supports nested Config.
        """
        d = {}
        for key, item in self._state.items():
            if isinstance(item, Config):
                d[key] = item.dict
            else:
                d[key] = item
        return d
    
    def keys(self):
        """List the config attributes.
        """
        return self._state.keys()

    def __contains__(self, key):
        return key in self._state

    def get(self, name, default):
        return self._state.get(name, default)
        
    def __getattr__(self, name):
        return self._state.get(name, None)

    def __setattr__(self, name, value):
        if name in ["_state", "_lock"]:
            object.__setattr__(self, name, value)
        else:
            if self._lock:
                raise Exception("Config is locked.")
            self._state[name] = value
            
    def __getstate__(self): 
        return self._state

    def __setstate__(self, d): 
        if self._lock == False:
            self._state.update(d)
        else:
            raise Exception("Config is locked.")
    
    def __getitem__(self, key):
        if key in self._state:
            return self._state[key]
        return None
        
    def __dir__(self):
        return self._state.keys()

    def __setitem__(self, key, value):
        if self._lock == False:
            self._state[key] = value
        else:
            raise Exception("Config is locked.")

    def __iter__(self):
        return self._state.__iter__()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return json.dumps(self.dict, indent=2)


from typing import List

class PaperContent():
    """Conveninent helper to write papers from the python-client.

    Usage
    --------------

    paper = PaperContent()
    paper.add('# Title')
    paper.add('## Subtitle')
    paper.add('Hello world')
    
    # Add image
    img_uid = run.add_artifact(path, wait_response=True)['uuid']
    paper.img(img_uid)
    
    # Push data
    paper.update_run_paper(run)

    """        
    def __init__(self):
        self.text = ''
    
    def add(self, text: str):
        self.text += text + '\n'
    
    def new_line(self):
        self.text += '\n'
        
    def img(self, uid: str):
        self.text += f'\n__ARTIFACT__[{uid}]\n\n'
        
    def imgs(self, uids: List[str]):
        self.text += '\n'
        for uid in uids:
            self.text += f'__ARTIFACT__[{uid}]\n'
        self.text += '\n'
    
    def hrule(self):
        self.text += '-------------------------------\n'
    
    def update_run_paper(self, run):
        run.set_paper(self.text)

    def __str__(self):
        return self.text
    
    def __repr__(self):
        return self.text


import os
import click
import getpass
import pyhectiqlab
import pyhectiqlab.ops as ops
from pyhectiqlab.utils import load_event_manager
from pyhectiqlab.auth import AuthProvider
from pyhectiqlab.mlmodels import download_mlmodel as ops_download_mlmodel
from pyhectiqlab.datasets import download_dataset as ops_download_dataset
from pyhectiqlab.apps import download_app as ops_download_app
from pyhectiqlab.apps import create_app as ops_create_app
import socket

@click.group()
def cli():
    """The official hectiqlab command line client"""
    pass

@cli.command()
def add_profile():
	auth = AuthProvider()
	username = click.prompt('Username ', )
	if auth.profile_exists(username):
		click.echo(f'A profile already exists for {username} at {auth.tokens_path}. Try using the pyhectiqlab or delete the profile before executing add-profile.')
		return

	password = getpass.getpass(prompt='Password: ', stream=None) 
	try:
		api_name = socket.gethostname()
	except:
		api_name = None

	name = click.prompt('Alias for the key: ', default=api_name)
	click.echo("Connecting...")
	success, api_key_uuid = auth.fetch_secret_api_token(username, password, name)

	if success:
		click.echo(f'Added profile [{username}] in {auth.tokens_path}')
		if (auth.profiles and len(auth.profiles)>1):
			click.echo(f"You have multiple profiles in {auth.tokens_path}. Use env var HECTIQLAB_PROFILE to switch between profiles.")
		click.echo('You can now start using hectiqlab 🎉!')
		click.echo('📚 Documentation at https://docs.hectiq.ai')
		click.echo('🚀 Web palteform at https://lab.hectiq.ai')

	else:
		click.echo('😞 Unsuccessful login.')

@cli.command()
def version():
	click.echo(pyhectiqlab.__version__)

@cli.command()
def profiles():
	auth = AuthProvider()
	profiles = list(auth.profiles.keys())
	click.echo(profiles)

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='MLModel name', help='Name of the mlmodel', required=True)
@click.option('-v', '--version', help='Version of the mlmodel. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=False)
@click.option('-o', '--overwrite', is_flag=True)
def download_mlmodel(project, name, version, save_path, overwrite):
	dir_path = ops_download_mlmodel(mlmodel_name=name, 
						project_path=project, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'MLModel saved in {dir_path}')

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='Dataset name', help='Name of the dataset', required=True)
@click.option('-v', '--version', help='Version of the dataset. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=True)
@click.option('-o', '--overwrite', is_flag=True)
def download_dataset(project, name, version, save_path, overwrite):
	dir_path = ops_download_dataset(dataset_name=name, 
						project_path=project, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'Dataset saved in {dir_path}')

@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='App name', help='Name of the app', required=True)
@click.option('-s', '--save_path', help='Save path [./]', default="./", required=True)
@click.option('-o', '--overwrite', is_flag=True)
@click.option('-r', '--raw', help='If set, the files are downloaded in save_path without directory', is_flag=True)
def download_app(project, name, save_path, overwrite, raw):
	dir_path = ops_download_app(app_name=name, 
						project_path=project, 
						save_path=save_path, 
						overwrite=overwrite,
						no_dir=raw)
	if dir_path is not None:
		click.echo(f'App saved in {dir_path}')


@cli.command()
@click.option('-p', '--project', prompt="Project", help='Path of the project (e.g. `hectiqai/project`)', required=True)
@click.option('-n', '--name', prompt='App name', help='Name of the app', required=True)
@click.option('-d', '--description', prompt='App description', help='A one-line description of what the app does', required=False)
@click.option('--files_path', type=click.Path(exists=True), prompt='Path to the folder containing the app', help='Path to the files',required=True, autocompletion=os.listdir)
@click.option('--privacy', prompt='Privacy settings (public or private)', type=click.Choice(['public', 'private']), required=True)
@click.option('--app_type', prompt='The app type',  help='The app type', type=click.Choice(['streamlit', 'gradio', 'plotly']), required=True)
def create_app(project, name, description, files_path, privacy, app_type):
	manager = load_event_manager()
	results = manager.add_event("fetch_instance_types", (project,), async_method=False)
	if results.get("status_code")!=200:
		click.echo(results.get("detail", "Something wrong occured."))
		return
	payload = results.get("payload")
	if payload.get("owned_by")=="organization":
		click.echo(f"Showing the instances for {payload.get('owner')}:")
	instance_types = results.get("results")
	choices = click.Choice([r["name"] for r in instance_types if r["selectable"]==True], case_sensitive=True)
	instances_description = "\n\n"
	for r in instance_types:
		available = ""
		if r["selectable"]==False:
			available = "Requies a team membership\n"
		instances_description += f"{available}[{r['name']}] {r['price']}\n  {r['description']}\n  {r['tip']}\n\n"

	instance_type = click.prompt(f"Select your instance type: {instances_description}", type=choices, show_choices=True)
	click.echo("Please wait while we create your app...")

	url = ops_create_app(name=name, 
				project=project,
				app_type=app_type, 
				instance=instance_type, 
				is_private=privacy=="private", 
				dir_path=files_path, 
				description=description)
	click.echo(f"🚀 Your app is not available at {url}")
	click.launch(url)